"""
Module :mod:`chebyshev_polynomials._class_interface`

This module implements a class based interface to the Chebyshev polynomials via the
class :class:`ChebyshevPolynomialBasis`.

"""

# === Imports ===

from typing import Literal, Optional, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray

from .._utils import (
    IntScalar,
    RealScalar,
    get_validated_alpha,
    get_validated_offset_along_axis,
    get_validated_order,
)
from ._func_interface import chebyshev_polyvander, get_validated_chebyshev_kind

# === Classes ===


class ChebyshevPolynomialBasis:
    """
    This class represents a basis of Chebyshev polynomials which are defined by their

    - order ``n``,
    - scaling factor ``alpha``,
    - center ``x_center`` (mu), and
    - kind (first or second kind).

    For their definition, please refer to the Notes section.

    Parameters
    ----------
    n : :class:`int`
        The order of the dilated Chebyshev polynomials.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Chebyshev polynomials.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.
    kind : {``1``, ``2``, ``"first"``, ``"second"``}, default=``"second"``
        The kind of Chebyshev polynomials to compute, which can either

        - ``1`` or ``"first"`` for the first kind,
        - ``2`` or ``"second"`` for the second kind.

    jit : :class:`bool`, default=``True``
        Whether to use the Numba-accelerated implementation (``True``) or the
        NumPy-based implementation (``False``).
        If Numba is not available, the function silently falls back to the NumPy-based
        implementation.

    Attributes
    ----------
    n : :class:`int`
        The order of the Chebyshev polynomials.
    alpha : :class:`float`
        The scaling factor of the independent variables.
    x_center : :class:`float`
        The x-center mu of the Chebyshev polynomials.
    kind : {``1``, ``2``}
        The kind of Chebyshev polynomials to compute.

    Methods
    -------
    __len__()
        Returns the number of Chebyshev basis functions that will be computed with the
        given parameters.
    eval(x, n, alpha, x_center, validate_parameters=True)
        Evaluates the Chebyshev polynomials at the given points with the specified
        parameters.
    __call__(x)
        Evaluates the Chebyshev polynomials at the given points with the class
        parameters.

    Raises
    ------
    TypeError
        If any of the input parameters is not of the expected type.
    ValueError
        If ``n`` is not a non-negative integer, ``alpha`` is not a positive number, or
        ``kind`` is not one of the allowed values.

    Notes
    -----
    The naive three-term recurrence relations for Chebyshev polynomials of the first and
    second kind have bad numerical properties because the error is increasing
    proportional to ``n * n``, i.e., quadratically with the order of the polynomial
    ``n``.
    However, a combined recurrence relation that links both kinds of Chebyshev
    polynomials has an error proportional to ``n``, i.e., a linear increase with the
    order of the polynomial ``n``. Already for ``n`` as little as ``4``, the combined
    recurrence relation outperforms the naive three-term recurrence relations. So, the
    roughly doubled computational cost of the combined recurrence relation is justified.
    On top of that, a combined evaluation offers the opportunity to easily compute the
    derivatives of the Chebyshev polynomials because the derivatives of the first kind
    are related to the polynomials of the second kind and vice versa.

    References
    ----------
    The implementation is taken from [1]_.

    .. [1] Hrycak T., Schmutzhard S., Accurate evaluation of Chebyshev polynomials in
       floating-point arithmetic, BIT Num

    """  # noqa: E501

    # --- Constructor ---

    def __init__(
        self,
        n: IntScalar,
        alpha: RealScalar = 1.0,
        x_center: Optional[RealScalar] = None,
        kind: Literal[1, 2, "first", "second"] = "second",
        jit: bool = True,
    ):
        self._n: int = get_validated_order(n=n)
        self._alpha: float = get_validated_alpha(alpha=alpha)
        self._x_center: float = get_validated_offset_along_axis(
            offset=x_center,
            which_axis="x",
        )
        self._kind: Literal[1, 2] = get_validated_chebyshev_kind(
            kind=kind,
            allow_both_kinds=False,
        )
        self._jit: bool = jit

    # --- Properties ---

    @property
    def n(self) -> int:
        return self._n

    @n.setter
    def n(self, value: IntScalar) -> None:
        self._n = get_validated_order(n=value)

    @property
    def alpha(self) -> float:
        return self._alpha

    @alpha.setter
    def alpha(self, value: RealScalar) -> None:
        self._alpha = get_validated_alpha(alpha=value)

    @property
    def x_center(self) -> float:
        return self._x_center

    @x_center.setter
    def x_center(self, value: Optional[RealScalar]) -> None:
        self._x_center = get_validated_offset_along_axis(
            offset=value,
            which_axis="x",
        )

    @property
    def kind(self) -> Literal[1, 2]:
        return self._kind

    @kind.setter
    def kind(self, value: Literal[1, 2, "first", "second"]) -> None:
        self._kind = get_validated_chebyshev_kind(
            kind=value,
            allow_both_kinds=False,
        )

    @property
    def jit(self) -> bool:
        return self._jit

    @jit.setter
    def jit(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(
                f"Expected 'jit' to be a boolean but it is of type "
                f"'{type(value).__name__}'."
            )

        self._jit = value

    # --- Public Methods ---

    @staticmethod
    def eval(
        x: Union[RealScalar, ArrayLike],
        n: IntScalar = 10,
        alpha: RealScalar = 1.0,
        x_center: Optional[RealScalar] = None,
        kind: Literal[1, 2, "first", "second"] = "second",
        jit: bool = True,
        validate_parameters: bool = True,
    ) -> NDArray[np.float64]:
        """
        Evaluates the Chebyshev polynomials at the given points with the specified
        parameters.

        Parameters
        ----------
        x : :class:`float` or :class:`int` or Array-like of shape (m,)
            The points at which the dilated Chebyshev polynomials are evaluated.
            Internally, it will be promoted to ``np.float64``.
            It has to contain at least one element.
        n : :class:`int`
            The order of the dilated Chebyshev polynomials.
            It must be a non-negative integer ``>= 0``.
        alpha : :class:`float` or :class:`int`, default=``1.0``
            The scaling factor of the independent variable ``x`` for
            ``x_scaled = x / alpha``.
            It must be a positive number ``> 0``.
        x_center : :class:`float` or :class:`int` or ``None``, default=``None``
            The center of the dilated Chebyshev polynomials.
            If ``None`` or ``0``, the functions are centered at the origin.
            Otherwise, the center is shifted to the given value.
        kind : {``1``, ``2``, ``"first"``, ``"second"``}, default=``"second"``
            The kind of Chebyshev polynomials to compute, which can either

            - ``1`` or ``"first"`` for the first kind,
            - ``2`` or ``"second"`` for the second kind.

        jit : :class:`bool`, default=``True``
            Whether to use the Numba-accelerated implementation (``True``) or the
            NumPy-based implementation (``False``).
            If Numba is not available, the function silently falls back to the
            NumPy-based implementation.
        validate_parameters : :class:`bool`, default=``True``
            Whether to validate ``n``, ``alpha``, and ``x_center`` before evaluating the
            Chebyshev polynomials.
            Disabling the checks is highly discouraged and was only implemented for
            internal purposes.

        Returns
        -------
        chebyshev_basis : :class:`numpy.ndarray` of shape (m, n + 1) of dtype ``np.float64``
            The values of the dilated Chebyshev polynomials at the points ``x``.
            It will always be 2D even if ``x`` is a scalar.

        Notes
        -----
        For a detailed description of the Chebyshev polynomials, please refer to the
        docstring of the class :class:`ChebyshevPolynomialBasis` itself.

        """  # noqa: E501

        return chebyshev_polyvander(
            x=x,
            n=n,
            alpha=alpha,
            x_center=x_center,
            kind=kind,
            allow_both_kinds=False,
            jit=jit,
            validate_parameters=validate_parameters,
        )

    # --- Magic Methods ---

    def __len__(self):
        """
        Returns the number of Chebyshev basis functions that will be computed, i.e.,
        ``n + 1``.

        """

        return self._n + 1

    def __call__(
        self,
        x: Union[RealScalar, ArrayLike],
    ) -> NDArray[np.float64]:
        """
        Evaluates the Chebyshev polynomials at the given points with the class
        parameters.
        Please refer to the docstrings of the class itself and the method :meth:`eval`
        for more information.

        """  # noqa: E501

        return self.eval(
            x=x,
            n=self._n,
            alpha=self._alpha,
            x_center=self._x_center,
            kind=self._kind,
            jit=self._jit,
            validate_parameters=False,
        )



"""
Module :mod:`_utils._validate`

This module implements the input validations for the Chebyshev polynomial and Hermite
function parameters:

- ``x``: the independent variable
- ``n``: the order
- ``alpha``: the scaling factor of the independent variable
- ``x_center``: the center

"""

# === Imports ===

from typing import Any, Tuple, Type, Union

import numpy as np

# === Types ===

# a Python or NumPy float or integer
RealScalar = Union[float, int, np.floating, np.integer]
_real_scalar_types_no_pyfloat = (int, np.floating, np.integer)

# a Python or NumPy integer
IntScalar = Union[int, np.integer]

# === Exceptions ===


class _GivesNoRealArrayError(Exception):
    """
    Exception raised when the input does not result in a real-valued NumPy array.

    """

    pass


# === Functions ===


def get_validated_grid_points(
    grid_points: Any,
    dtype: Type,
    name: str = "x",
) -> np.ndarray:
    """
    Validates the input for grid points and returns the validated input.

    """

    # the grid points are converted to a 1D NumPy array for checking
    try:
        grid_points_internal = np.atleast_1d(grid_points)
        if not np.isreal(grid_points_internal).all():
            raise _GivesNoRealArrayError()

    except _GivesNoRealArrayError:
        grid_points_type_str = f"{type(grid_points)}"
        if hasattr(grid_points, "dtype"):  # pragma: no cover
            grid_points_type_str += f" with dtype {grid_points.dtype}"  # type: ignore

        raise TypeError(
            f"Expected '{name}' to be a real scalar or a real-value Array-like but got "
            f"type {grid_points_type_str}."
        )

    # if the converted grid points are not 1-dimensional, an error is raised
    if grid_points_internal.ndim != 1:
        raise ValueError(
            f"Expected '{name}' to be 1-dimensional but it is "
            f"{grid_points_internal.ndim}-dimensional."
        )

    # if required, the dtype is converted to the target dtype
    if not grid_points_internal.dtype == dtype:
        grid_points_internal = grid_points_internal.astype(dtype)

    if grid_points_internal.size < 1:
        raise ValueError(
            f"Expected '{name}' to have at least one element but got "
            f"{grid_points_internal.size} elements."
        )

    return grid_points_internal


def get_validated_order(n: Any) -> int:
    """
    Validates the input for the order of the Chebyshev polynomials or Hermite functions
    and returns the validated input.

    """

    # NumPy integers need to be converted to Python integers
    if isinstance(n, np.integer):
        n = int(n)

    if not isinstance(n, int):
        raise TypeError(f"Expected 'n' to be an integer but got type {type(n)}.")

    if n < 0:
        raise ValueError(f"Expected 'n' to be a non-negative integer but got {n}.")

    return n


def get_validated_alpha(alpha: Any) -> float:
    """
    Validates the input for the scaling factor for the independent variable of the
    Chebyshev polynomials or Hermite functions and returns the validated input.

    """

    # integers and NumPy scalars need to be converted to Python floats
    if isinstance(alpha, _real_scalar_types_no_pyfloat):
        alpha = float(alpha)

    if not isinstance(alpha, float):
        raise TypeError(
            f"Expected 'alpha' to be a float or integer but got type {type(alpha)}."
        )

    if alpha <= 0.0:
        raise ValueError(f"Expected 'alpha' to be a positive number but got {alpha}.")

    return alpha


def get_validated_offset_along_axis(
    offset: Any,
    which_axis: str,
) -> float:
    """
    Validates the input for the center (shift) along the an axis, e.g., the x-center
    along the x-axis, and returns the validated input.

    """

    if offset is not None:
        # integers and NumPy scalars need to be converted to Python floats
        if isinstance(offset, _real_scalar_types_no_pyfloat):
            offset = float(offset)

        if not isinstance(offset, float):
            raise TypeError(
                f"Expected the {which_axis}-'center' to be a float, integer, or None "
                f"but got type {type(offset)}."
            )

    else:
        offset = 0.0

    return offset


def get_validated_chebpoly_or_hermfunc_input(
    x: Any,
    x_dtype: Type,
    n: Any,
    alpha: Any,
    x_center: Any,
) -> Tuple[np.ndarray, int, float, float]:
    """
    Validates the input for the Chebyshev polynomials or Hermite functions and returns
    the validated input.

    """

    # the input is validated according to the requirements of the higher level caller
    # functions
    return (
        get_validated_grid_points(grid_points=x, dtype=x_dtype, name="x"),
        get_validated_order(n=n),
        get_validated_alpha(alpha=alpha),
        get_validated_offset_along_axis(
            offset=x_center,
            which_axis="x",
        ),
    )



"""
Module :mod:`hermite_functions._approximations`

This module offers approximation functions for special points of the Hermite functions,
namely

- the x-position of their largest zero (= outermost root where y = 0)
- the x-position at which the outermost tail fades below machine precision
- the x- and y-position of the maximum of the Hermite functions in their outermost
    oscillation

"""

# === Imports ===

from math import log as pylog
from math import sqrt as pysqrt
from typing import Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.interpolate import splev

from .._utils import (
    IntScalar,
    RealScalar,
    get_validated_alpha,
    get_validated_grid_points,
    get_validated_offset_along_axis,
    get_validated_order,
)
from ._hermite_largest_extrema_spline import (
    HERMITE_LARGEST_EXTREMA_MAX_ORDER,
    X_HERMITE_LARGEST_EXTREMA_SPLINE_TCK,
    Y_HERMITE_LARGEST_EXTREMA_SPLINE_TCK,
)
from ._hermite_largest_roots_spline import (
    HERMITE_LARGEST_ZEROS_MAX_ORDER,
    HERMITE_LARGEST_ZEROS_SPLINE_TCK,
)
from ._hermite_tail_gauss_sigma_spline import (
    HERMITE_TAIL_GAUSS_APPROX_MAX_ORDER,
    HERMITE_TAIL_GAUSS_SIGMA_SPLINE_TCK,
)

# === Constants ===

# the logarithm of the machine precision for float64
_LOG_DOUBLE_EPS = pylog(np.finfo(np.float64).eps)


# === Models ===


# a class that represents a squared exponential peak that can be evaluated at a given
# x-position


class TailSquaredExponentialApprox:
    """
    A class that represents a squared exponential or scaled Gaussian peak to approximate
    the outermost tail of the Hermite functions. It can be evaluated at a given
    x-position and also solve for the x-position at which the peak reaches a given
    y-value.

    Parameters
    ----------
    center_mu : :class:`float` or :class:`int`
        The center at which the y-value of the peak is ``amplitude``.
    stddev_sigma : :class:`float` or :class:`int`
        The standard deviation of the peak.
    amplitude : :class:`float` or :class:`int`
        The amplitude of the peak that is reached at the ``center_mu``.
    side: ``{"left", "right"}``
        The side of the Hermite function for which the peak is approximated.
        The fadeout point of the Hermite functions will be located at this side from
        the ``center_mu``.

    Attributes
    ----------
    center_mu, stddev_sigma, amplitude : :class:`float`
        The center, standard deviation, and amplitude of the peak.
    side : ``{"left", "right"}``
        The side of the Hermite function for which the peak is approximated.

    Methods
    -------
    __call__(x)
        Evaluates the peak at the given x-position.
    solve_for_y_fraction(y_fraction)
        Solves for the x-position at which the peak reaches the given y-value as a
        fraction of the amplitude.

    """

    def __init__(
        self,
        center_mu: RealScalar,
        stddev_sigma: RealScalar,
        amplitude: RealScalar,
        side: Literal["left", "right"],
    ) -> None:
        self.center_mu: RealScalar = center_mu
        self.stddev_sigma: RealScalar = stddev_sigma
        self.amplitude: RealScalar = amplitude
        self.side: Literal["left", "right"] = side

    def __call__(
        self,
        x: Union[RealScalar, ArrayLike],
    ) -> Union[float, np.ndarray]:
        """
        Evaluates the peak at the given x-position.

        Parameters
        ----------
        x : :class:`float` or :class:`int` or Array-like of shape (m,)
            The points at which the peak is evaluated.
            Internally, it will be promoted to ``np.float64``.
            It has to contain at least one element.

        Returns
        -------
        peak_values : :class:`float` or :class:`numpy.ndarray` of shape (m,)
            The y-values of the peak at the given x-positions.
            For scalar ``x``, the result is a also a scalar while for array-like ``x``,
            the result is an Array as well.

        Raises
        ------
        TypeError
            If ``x`` is not of the expected type.
        ValueError
            If ``x`` is not 1-dimensional after conversion to a NumPy array.

        """
        # the x-values are validated
        x_internal = get_validated_grid_points(grid_points=x, dtype=np.float64)

        # the peak is evaluated at the given x-values
        peak_values = self.amplitude * np.exp(
            np.negative(
                0.5 * np.square((x_internal - self.center_mu) / self.stddev_sigma)
            )
        )

        if np.isscalar(x):
            return float(peak_values[0])

        return peak_values

    def solve_for_y_fraction(
        self,
        y_fraction: Union[RealScalar, ArrayLike],
    ) -> Union[float, np.ndarray]:
        """
        Solves for the x-position at which the peak reaches the given y-value as a
        fraction of the amplitude.
        For a ``"left"`` side peak, the x-position will be the leftmost of both possible
        solutions while for a ``"right"`` side peak, it will be the rightmost.

        Parameters
        ----------
        y_fraction : :class:`float` or :class:`int` or Array-like of shape (m,)
            The y-values for which the x-positions are solved as a fraction of the
            amplitude.
            Internally, it will be promoted to ``np.float64``.
            It has to contain at least one element.

        Returns
        -------
        x_positions : :class:`float` or :class:`numpy.ndarray` of shape (m,)
            The x-positions at which the peak reaches the fraction of the amplitude.
            It will be the left or right solution for the ``"left"`` or ``"right"``
            side, respectively.
            For scalar ``y``, the result is a also a scalar while for array-like ``y``,
            the result is an Array as well.
            It will contain ``NaN``-values for y-values that are not in the range of the
            peak.

        Raises
        ------
        TypeError
            If ``y`` is not of the expected type.
        ValueError
            If ``y`` is not 1-dimensional after conversion to a NumPy array.

        """

        # the y-values are validated
        y_internal = get_validated_grid_points(
            grid_points=y_fraction,
            dtype=np.float64,
            name="y_fraction",
        )

        # the x-positions are solved for the given y-values
        sign = -1.0 if self.side == "left" else 1.0
        x_positions = self.center_mu + sign * self.stddev_sigma * np.sqrt(
            -2.0 * np.log(y_internal)
        )

        if np.isscalar(y_fraction):
            return float(x_positions[0])

        return x_positions


# === Auxiliary Functions ===


def _apply_centering_and_scaling(
    values: np.ndarray,
    alpha: float,
    x_center: float,
) -> np.ndarray:
    """
    Applies centering and scaling to the given values.

    """

    # the scaling is applied (if required)
    if alpha != 1.0:
        values *= alpha

    # the centering is applied (if required)
    if x_center != 0.0:
        values += x_center

    return values


# === Functions ===


def x_largest_zeros(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> np.ndarray:
    """
    Approximates the x-position of the largest zero (= outermost root) of the Hermite
    functions for a given order and scaling factor.
    Please refer to the Notes for further details on the approximation.

    Parameters
    ----------
    n : :class:`int`
        The order of the Hermite functions.
        It must be a non-negative integer ``>= 0`` and less than or equal to the maximum
        order for the spline interpolation (roughly 100 000).
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    x_largest_zeros : :class:`numpy.ndarray` of shape (0,) or (1,) or (2,)
        The x-positions of the left and right largest zeros of the Hermite functions
        (in that order) for the given order, scaling factor, and center.
        It will hold two entries except for

        - order 0 where there is no zero and
        - order 1 where the zero is exactly at ``x_center``.

    Raises
    ------
    NotImplementedError
        If the order is larger than the maximum order for the spline interpolation.

    Notes
    -----
    The approximation is based on a spline that mimics the results of
    ``scipy.special.roots_hermite``, but this function gives roots whose y-values are in
    the order of ``1e-8``. Thus, the reference values are the limiting factor for
    accuracy and not the spline itself.
    Given that it's based on a spline, the approximation is limited to a maximum order
    of roughly 100 000, but this should be sufficient for most applications.

    """

    # --- Input Validation ---

    n = get_validated_order(n=n)
    alpha = get_validated_alpha(alpha=alpha)
    x_center = get_validated_offset_along_axis(offset=x_center, which_axis="x")

    # --- Computation ---

    # if the order exceeds the maximum order for the spline interpolation, an error is
    # raised
    if n > HERMITE_LARGEST_ZEROS_MAX_ORDER:  # pragma: no cover
        raise NotImplementedError(
            f"Order {n} exceeds the maximum order {HERMITE_LARGEST_ZEROS_MAX_ORDER} "
            f"for the spline interpolation for the largest zeros."
        )

    # if the order exceeds 1, the spline interpolation is used to approximate the
    # largest zeros
    if n > 1:
        x_largest_zero_positive = splev(
            x=n,
            tck=HERMITE_LARGEST_ZEROS_SPLINE_TCK,
        )
        x_largest_zeros = np.array(
            [
                -x_largest_zero_positive,  # type: ignore
                x_largest_zero_positive,
            ]
        )

    # for order 1, the zero is exactly at 0
    elif n == 1:
        x_largest_zeros = np.array([0.0])

    # for order 0, there is no zero
    else:
        x_largest_zeros = np.array([])

    # the centering and scaling are applied
    return _apply_centering_and_scaling(
        values=x_largest_zeros,
        alpha=alpha,
        x_center=x_center,
    )


def x_largest_extrema(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> np.ndarray:
    """
    Approximates the x-position of the maximum of the Hermite functions in their
    outermost oscillation for a given order and scaling factor.
    Please refer to the Notes for further details on the approximation.

    Parameters
    ----------
    n : :class:`int`
        The order of the Hermite functions.
        It must be a non-negative integer ``>= 0`` and less than or equal to the maximum
        order for the spline interpolation (roughly 100 000).
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    x_largest_extrema : :class:`numpy.ndarray` of shape (1,) or (2,)
        The x-positions of the maximum of the Hermite functions in their outermost
        oscillation (in that order) for the given order, scaling factor, and center.
        It will hold two entries except for

        - order 0 where the maximum is exactly at ``x_center``.

    Raises
    ------
    NotImplementedError
        If the order is larger than the maximum order for the spline interpolation.

    Notes
    -----
    The approximation is based on a spline interpolation of the results of a numerical
    optimisation to find the maximum of the Hermite functions.
    Given that it's based on a spline, the approximation is limited to a maximum order
    of roughly 100 000, but this should be sufficient for most applications.

    """

    # --- Input Validation ---

    n = get_validated_order(n=n)
    alpha = get_validated_alpha(alpha=alpha)
    x_center = get_validated_offset_along_axis(offset=x_center, which_axis="x")

    # --- Computation ---

    # if the order exceeds the maximum order for the spline interpolation, an error is
    # raised
    if n > HERMITE_LARGEST_EXTREMA_MAX_ORDER:  # pragma: no cover
        raise NotImplementedError(
            f"Order {n} exceeds the maximum order {HERMITE_LARGEST_EXTREMA_MAX_ORDER} "
            f"for the spline interpolation for the largest extrema."
        )

    # if the order is not 0, the spline interpolation is used to approximate the
    # largest extrema
    if n > 0:
        x_largest_extrema_positive = splev(
            x=n,
            tck=X_HERMITE_LARGEST_EXTREMA_SPLINE_TCK,
        )
        x_largest_extrema = np.array(
            [
                -x_largest_extrema_positive,  # type: ignore
                x_largest_extrema_positive,
            ]
        )

    # for order 0, the maximum is exactly at the center
    else:
        x_largest_extrema = np.array([0.0])

    # the centering and scaling are applied
    return _apply_centering_and_scaling(
        values=x_largest_extrema,
        alpha=alpha,
        x_center=x_center,
    )


def y_largest_extrema(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> np.ndarray:
    """
    Approximates the y-position of the maximum of the Hermite functions in their
    outermost oscillation for a given order and scaling factor.
    Please refer to the Notes for further details on the approximation.

    Parameters
    ----------
    n : :class:`int`
        The order of the Hermite functions.
        It must be a non-negative integer ``>= 0`` and less than or equal to the maximum
        order for the spline interpolation (roughly 100 000).
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.
        This value has no effect on the result.

    Returns
    -------
    y_largest_extrema : :class:`numpy.ndarray` of shape (1,) or (2,)
        The y-positions of the maximum of the Hermite functions in their outermost
        oscillation (in that order) for the given order, scaling factor, and center.
        It will hold two entries except for

        - order 0 where the maximum is exactly at
            :math:`\\frac{1}{\\sqrt[4]{\\pi\\cdot\\alpha^{2}}}`

    Raises
    ------
    NotImplementedError
        If the order is larger than the maximum order for the spline interpolation.

    Notes
    -----
    The approximation is based on a spline interpolation of the results of a the
    evaluation of the Hermite functions at the x-position provided by
    :func:`x_largest_extrema`.

    """

    # --- Input Validation ---

    n = get_validated_order(n=n)
    alpha = get_validated_alpha(alpha=alpha)

    # --- Computation ---

    # if the order exceeds the maximum order for the spline interpolation, an error is
    # raised
    if n > HERMITE_LARGEST_EXTREMA_MAX_ORDER:  # pragma: no cover
        raise NotImplementedError(
            f"Order {n} exceeds the maximum order {HERMITE_LARGEST_EXTREMA_MAX_ORDER} "
            f"for the spline interpolation for the largest extrema."
        )

    # if the order is exactly zero, the maximum is returned directly
    if n == 0:
        return np.array([1.0 / pysqrt(pysqrt(np.pi * alpha * alpha))])

    # otherwise, the spline is evaluated to get the y-position of the maximum
    y_largest_extrema_positive = splev(
        x=n,
        tck=Y_HERMITE_LARGEST_EXTREMA_SPLINE_TCK,
    )
    # it still needs to be scaled by the square root of alpha
    y_largest_extrema_positive /= pysqrt(alpha)  # type: ignore

    # for even orders, both extrema are the same and positive
    if n % 2 == 0:
        return np.array([y_largest_extrema_positive, y_largest_extrema_positive])

    # for odd orders, the extrema are the same, but the left one is positive while the
    # right one is negative
    return np.array(
        [
            -y_largest_extrema_positive,
            y_largest_extrema_positive,
        ]
    )


def x_and_y_largest_extrema(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Approximates the x- and y-positions of the maximum of the Hermite functions in their
    outermost oscillation for a given order and scaling factor.
    It is a convenience function that combines the results of :func:`x_largest_extrema`
    and :func:`y_largest_extrema`.

    Parameters
    ----------
    n : :class:`int`
        The order of the Hermite functions.
        It must be a non-negative integer ``>= 0`` and less than or equal to the maximum
        order for the spline interpolation (roughly 100 000).
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    x_largest_extrema, y_largest_extrema : :class:`numpy.ndarray` of shape (1,) or (2,)
        The x- and y-positions of the maximum of the Hermite functions in their outermost
        oscillation (in that order) for the given order, scaling factor, and center.
        They will hold two entries each except for

        - order 0 where the maximum is exactly at the x-position ``x_center`` and the
            y-position is :math:`\\frac{1}{\\sqrt[4]{\\pi\\cdot\\alpha^{2}}}`

    Raises
    ------
    NotImplementedError
        If the order is larger than the maximum order for the spline interpolations.

    """  # noqa: E501

    return (
        x_largest_extrema(
            n=n,
            alpha=alpha,
            x_center=x_center,
        ),
        y_largest_extrema(
            n=n,
            alpha=alpha,
            x_center=x_center,
        ),
    )


def x_fadeout(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> np.ndarray:
    """
    Approximates the x-position at which the outermost tail of the dilated Hermite
    functions drops below machine precision when compared to the maximum value for
    ``float64``.
    Please refer to the Notes for further details on the approximation.

    Parameters
    ----------
    n : :class:`int`
        The order of the dilated Hermite functions.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    x_fadeouts : :class:`numpy.ndarray` of shape (2,)
        The x-positions of the left and right fadeouts of the outermost tail of the
        Hermite functions (in that order) for the given order, scaling factor, and
        center.

    Notes
    -----
    At the fadeout point, there is no point in a complex integral representation of the
    Hermite functions that adds any numerically non-zero contribution. This
    approximation might be a bit conservative, but it provides consistent fadeout
    points.

    References
    ----------
    The implementation is derived from [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    """

    # --- Input Validation ---

    n = get_validated_order(n=n)
    alpha = get_validated_alpha(alpha=alpha)
    x_center = get_validated_offset_along_axis(offset=x_center, which_axis="x")

    # --- Computation ---

    # for orders > 0, the gamma-value is computed which is required for the estimation
    if n > 0:
        sqrt_n = pysqrt(n)
        k_value = 0.7071067811865476 * sqrt_n  # k = sqrt(n / 2)
        log_n_factorial = np.log(  # log(n!)
            np.arange(
                start=1,
                stop=max(2, n + 1),
                step=1,
                dtype=np.int64,
            )
        ).sum()
        log_gamma = (
            0.5 * log_n_factorial
            - 0.5 * n * pylog(2.0)  # log(2 ** (n / 2))
            - n * pylog(k_value)  # log(k ** n)
            - 0.25 * pylog(np.pi)  # log(pi ** 0.25)
        )

        # then, the approximation is made by finding the point where the arccosine of
        # the argument of the complex integral representation of the Hermite functions
        # exceeds 1.0; beyond this point, the contribution of any part of the complex
        # integral representation is numerically zero
        x_fadeout_positive = pysqrt(2.0) * sqrt_n + pysqrt(
            n + 2.0 * (log_gamma - _LOG_DOUBLE_EPS - pylog(6.5 * sqrt_n))
        )

    # for order 0, the fadeout is taken as the point where the Gaussian term drops
    # below machine precision
    else:
        # the argument of the Gaussian term is -0.5 * x ** 2, and the fadeout is
        # computed by solving for the point where the exponent exceeds the logarithm of
        # the machine precision
        x_fadeout_positive = pysqrt(-2.0 * _LOG_DOUBLE_EPS)

    # finally, the fadeouts are stored as an Array
    x_fadeouts = np.array(
        [
            -x_fadeout_positive,  # type: ignore
            x_fadeout_positive,
        ]
    )

    # the centering and scaling are applied
    return _apply_centering_and_scaling(
        values=x_fadeouts,
        alpha=alpha,
        x_center=x_center,
    )


def get_tail_gauss_fit(
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> Tuple[TailSquaredExponentialApprox, TailSquaredExponentialApprox]:
    """
    Approximates the outermost tail of the dilated Hermite functions with a squared
    exponential or scaled Gaussian peak.
    This is a very crude approximation that is only valid for the respective side that
    goes towards the fadeout point. One of its characteristics is that the Gaussian peak
    is too wide in most of the points to get conservative estimates for the tail
    behaviour.

    Parameters
    ----------
    n : :class:`int`
        The order of the dilated Hermite functions.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    left_tail, right_tail : :class:`TailSquaredExponentialApprox`
        The squared exponential or scaled Gaussian peaks that approximate the outermost
        tail of the Hermite functions on the left and right side of the maximum,
        respectively.
        They can be called directly to evaluate the peak at a given x-position or to
        solve for the x-position at which the peak reaches a given y-value.
        Please refer to the documentation of the class
        :class:`TailSquaredExponentialApprox` for further details.

    Raises
    ------
    NotImplementedError
        If the order is larger than the maximum order for the spline interpolations.

    """

    # if the order is larger than the maximum order for the spline interpolation of the
    # Gaussian standard deviation, an error is raised
    if n > HERMITE_TAIL_GAUSS_APPROX_MAX_ORDER:  # pragma: no cover
        raise NotImplementedError(
            f"Order {n} exceeds the maximum order "
            f"{HERMITE_TAIL_GAUSS_APPROX_MAX_ORDER} for the spline interpolation for "
            f"the tail Gaussian approximation."
        )

    # first, the xy-coordinates of the largest extrema are computed
    # NOTE: this also performs the input validation
    x_extrema, y_extrema = x_and_y_largest_extrema(
        n=n,
        alpha=alpha,
        x_center=x_center,
    )

    # if the order is 0, then there would only be one fit that fits perfectly over the
    # whole curve; yet 2 peaks are required, so the extrema are simply repeated
    if n == 0:
        x_extrema = np.repeat(x_extrema, repeats=2)
        y_extrema = np.repeat(y_extrema, repeats=2)

    # afterwards, the standard deviation sigma is computed by evaluating the
    # pre-computed spline curve at the natural logarithm of the order
    # the spline is only valid for nonzero orders
    if n > 0:
        stddev_sigma = splev(
            x=pylog(n),
            tck=HERMITE_TAIL_GAUSS_SIGMA_SPLINE_TCK,
        )
        # for proper scaling, the standard deviation is multiplied with the alpha value
        stddev_sigma *= alpha  # type: ignore

    # for order zero, the standard deviation is simply the alpha value
    else:
        stddev_sigma = alpha

    # the peaks are created
    left_tail = TailSquaredExponentialApprox(
        center_mu=x_extrema[0],
        stddev_sigma=stddev_sigma,  # type: ignore
        amplitude=y_extrema[0],
        side="left",
    )
    right_tail = TailSquaredExponentialApprox(
        center_mu=x_extrema[1],
        stddev_sigma=stddev_sigma,  # type: ignore
        amplitude=y_extrema[1],
        side="right",
    )

    return left_tail, right_tail


def x_tail_drop_to_fraction(
    n: IntScalar,
    y_fraction: Union[RealScalar, ArrayLike],
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
) -> np.ndarray:
    """
    Approximates the x-position at which the outermost tail of the dilated Hermite
    functions drops below a given y-value as a fraction of the maximum value.
    Please refer to the Notes for further details on the approximation.

    Parameters
    ----------
    n : :class:`int`
        The order of the dilated Hermite functions.
        It must be a non-negative integer ``>= 0``.
    y_fraction : :class:`float` or :class:`int` or Array-like of shape (m,)
        The y-values for which the x-position is solved as a fraction of the maximum
        value that the ``n``-th order Hermite function reaches.
        Each value will have 2 solutions, one for the left and one for the right side.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.

    Returns
    -------
    x_drop : :class:`numpy.ndarray` of shape (m, 2)
        The x-positions of the left and right points at which the outermost tail of the
        Hermite functions drops below the given ``y_fraction``.
        Its first column corresponds to the left while the second column corresponds to
        the right solution, even if ``y_fraction`` is a scalar.

    Notes
    -----
    The approximation is based on the Gaussian tail approximation that is computed by
    :func:`get_tail_gauss_fit`. The x-positions are then solved for the given
    y-value as a fraction of the maximum value.

    """

    # the Gaussian fit of the tails is obtained
    left_tail, right_tail = get_tail_gauss_fit(
        n=n,
        alpha=alpha,
        x_center=x_center,
    )

    # the x-positions are solved for the given y-value
    return np.column_stack(
        (
            left_tail.solve_for_y_fraction(y_fraction=y_fraction),
            right_tail.solve_for_y_fraction(y_fraction=y_fraction),
        )
    )



"""
Module :mod:`hermite_functions`

This module provides  implementations of the Hermite functions which are essential for
the robust Fourier transform.
The ``n``-th dilated and shifted Hermite function is given by

.. image:: docs/hermite_functions/equations/HF-01-Hermite_Functions_TimeSpace_Domain.svg

where :math:`H_{n}` is the :math:`n`-th Hermite polynomial

.. image:: docs/hermite_functions/equations/HF-02-Hermite_Polynomials_TimeSpace_Domain.svg

They have two nice properties:
- they are orthogonal
- they are eigenfunctions of the Fourier transform which makes them suitable for the
    robust Fourier transform by Least Squares Fits.

"""  # noqa: E501

# === Imports ===

from . import _approximations as hermite_approx  # noqa: F401
from ._class_interface import HermiteFunctionBasis  # noqa: F401
from ._func_interface import (  # noqa: F401
    hermite_function_vander,
    single_hermite_function,
)



"""
Module :mod:`chebyshev_polynomials._func_interface`

This module implements the interface to the either NumPy-based or Numba-based
implementations of the Chebyshev polynomials of the first and second kind.

It augments them with an additional input validation which is better done in Python
and also handles the incorporation of the Numba-functions if Numba is available at
runtime.

"""

# === Imports ===

from typing import Any, Literal, Optional, Tuple, Union, overload

import numpy as np
from numpy.typing import ArrayLike, NDArray

from .._utils import (
    IntScalar,
    RealScalar,
    get_validated_chebpoly_or_hermfunc_input,
    get_validated_grid_points,
    normalise_x_values,
)
from ._numba_funcs import nb_chebyshev_polyvander as _nb_chebyshev_polyvander
from ._numpy_funcs import _chebyshev_polyvander as _np_chebyshev_polyvander

# === Constants ===

# the set of kinds for the first and second kind of Chebyshev polynomials
chebyshev_first_kind_set = {1, "first"}
chebyshev_second_kind_set = {2, "second"}
chebyshev_both_kinds_set = {"both", None}


# === Auxiliary Functions ===


@overload
def get_validated_chebyshev_kind(
    kind: Any,
    allow_both_kinds: Literal[False],
) -> Literal[1, 2]: ...


@overload
def get_validated_chebyshev_kind(
    kind: Any,
    allow_both_kinds: Literal[True] = True,
) -> Optional[Literal[1, 2]]: ...


@overload
def get_validated_chebyshev_kind(
    kind: Any,
    allow_both_kinds: bool = True,
) -> Optional[Literal[1, 2]]: ...


def get_validated_chebyshev_kind(
    kind: Any,
    allow_both_kinds: bool = True,
) -> Optional[Literal[1, 2]]:
    """
    Validates the input for the kind of Chebyshev polynomials and returns the validated
    input.

    """

    # if the kind is not of the expected type, an error is raised
    if not isinstance(kind, (int, str)) and kind is not None:
        raise TypeError(
            f"Expected 'kind' to be an integer, a string, or None, but got type "
            f"{type(kind)}."
        )

    # then, the kind is checked for being one of the allowed values
    kind_internal = kind
    if isinstance(kind, str):
        kind_internal = kind.lower()

    if kind_internal in chebyshev_first_kind_set:
        return 1

    if kind_internal in chebyshev_second_kind_set:
        return 2

    if kind_internal in chebyshev_both_kinds_set:
        if allow_both_kinds:
            return None

        raise ValueError(
            f"Expected 'kind' to be one of {chebyshev_first_kind_set} or "
            f"{chebyshev_second_kind_set}, but got '{kind}'."
        )

    raise ValueError(
        f"Expected 'kind' to be one of {chebyshev_first_kind_set}, "
        f"{chebyshev_second_kind_set}, or {chebyshev_both_kinds_set}, but got "
        f"'{kind}'."
    )


# === Main Functions ===


@overload
def chebyshev_polyvander(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    allow_both_kinds: bool = True,
    jit: bool = True,
    validate_parameters: bool = True,
    *,
    kind: Literal[1, 2, "first", "second"] = "second",
) -> NDArray[np.float64]: ...


@overload
def chebyshev_polyvander(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    allow_both_kinds: bool = True,
    jit: bool = True,
    validate_parameters: bool = True,
    *,
    kind: Literal["both"],
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]: ...


@overload
def chebyshev_polyvander(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    allow_both_kinds: bool = True,
    jit: bool = True,
    validate_parameters: bool = True,
    *,
    kind: None,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]: ...


def chebyshev_polyvander(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    allow_both_kinds: bool = True,
    jit: bool = True,
    validate_parameters: bool = True,
    *,
    kind: Optional[Literal[1, 2, "first", "second", "both"]] = "second",
) -> Union[NDArray[np.float64], Tuple[NDArray[np.float64], NDArray[np.float64]]]:
    """
    Computes the basis (Vandermonde matrix) of dilated Chebyshev polynomials up to order
    ``n`` for the given points ``x``. It makes use of a combined recursion formula that
    links the first and second kind of Chebyshev polynomials to evaluate them in a
    numerically stable way.

    Parameters
    ----------
    x : :class:`float` or :class:`int` or Array-like of shape (m,)
        The points at which the dilated Chebyshev polynomials are evaluated.
        Internally, it will be promoted to ``np.float64``.
        It has to contain at least one element.
    n : :class:`int`
        The order of the dilated Chebyshev polynomials.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Chebyshev polynomials.
        If ``None`` or ``0``, the functions are centered at the origin.
        Otherwise, the center is shifted to the given value.
    kind : {``1``, ``2``, ``"first"``, ``"second"``, ``"both"``} or None, default=``"second"``
        The kind of Chebyshev polynomials to compute, which can either

        - ``1`` or ``"first"`` for the first kind,
        - ``2`` or ``"second"`` for the second kind,
        - ``"both"`` or ``None`` for both kinds simultaneously (no significant
            performance impact due to the combined recursion formula; only available if
            ``allow_both_kinds`` is ``True``).

        This is a keyword-only argument.

    allow_both_kinds : :class:`bool`, default=``True``
        Whether to allow the computation of both kinds of Chebyshev polynomials
        simultaneously (``True``) or not (``False``).
    jit : :class:`bool`, default=``True``
        Whether to use the Numba-accelerated implementation (``True``) or the
        NumPy-based implementation (``False``).
        If Numba is not available, the function silently falls back to the NumPy-based
        implementation.
    validate_parameters : :class:`bool`, default=``True``
        Whether to validate all the input parameters (``True``) or only ``x``
        (``False``).
        Disabling the input checks is not recommended and was only implemented for
        internal use.

    Returns
    -------
    chebyshev_t1_vander : :class:`numpy.ndarray` of shape (m, n + 1)
        The values of the Chebyshev polynomials of the first kind evaluated at the
        points ``x`` represented as a Vandermonde matrix.
        It will always be 2D even if ``x`` is a scalar.
        It is only returned if ``kind`` is ``1``, ``"first"``, ``"both"`` (for the
        latter, it is the first element of the tuple).

    chebyshev_u2_vander : :class:`numpy.ndarray` of shape (m, n + 1)
        The values of the Chebyshev polynomials of the second kind evaluated at the
        points ``x`` represented as a Vandermonde matrix.
        It will always be 2D even if ``x`` is a scalar.
        It is only returned if ``kind`` is ``2``, ``"second"``, ``"both"`` (for the
        latter, it is the second element of the tuple).

    Raises
    ------
    TypeError
        If any of the input arguments is not of the expected type.
    ValueError
        If ``x`` is not 1-dimensional after conversion to a NumPy array.
    ValueError
        If ``n`` is not a non-negative integer or ``alpha`` is not a positive number.
    ValueError
        If ``kind`` is not one of the allowed values.

    Notes
    -----
    The naive three-term recurrence relations for Chebyshev polynomials of the first and
    second kind have bad numerical properties because the error is increasing
    proportional to ``n * n``, i.e., quadratically with the order of the polynomial
    ``n``.
    However, a combined recurrence relation that links both kinds of Chebyshev
    polynomials has an error proportional to ``n``, i.e., a linear increase with the
    order of the polynomial ``n``. Already for ``n`` as little as ``4``, the combined
    recurrence relation outperforms the naive three-term recurrence relations. So, the
    roughly doubled computational cost of the combined recurrence relation is justified.
    On top of that, a combined evaluation offers the opportunity to easily compute the
    derivatives of the Chebyshev polynomials because the derivatives of the first kind
    are related to the polynomials of the second kind and vice versa.

    References
    ----------
    The implementation is taken from [1]_.

    .. [1] Hrycak T., Schmutzhard S., Accurate evaluation of Chebyshev polynomials in
       floating-point arithmetic, BIT Num

    """  # noqa: E501

    # --- Input validation ---

    if validate_parameters:
        (
            x_internal,
            n,
            alpha,
            x_center,
        ) = get_validated_chebpoly_or_hermfunc_input(
            x=x,
            x_dtype=np.float64,
            n=n,
            alpha=alpha,
            x_center=x_center,
        )

    else:  # pragma: no cover
        x_internal = get_validated_grid_points(grid_points=x, dtype=np.float64)

    kind_internal = get_validated_chebyshev_kind(
        kind=kind,
        allow_both_kinds=allow_both_kinds,
    )

    # --- Computation ---

    # if required, the x-values are centered
    x_internal = normalise_x_values(
        x_internal=x_internal,
        x=x,
        x_center=x_center,  # type: ignore
        alpha=alpha,  # type: ignore
    )

    # if requested, the Numba-accelerated implementation is used
    # NOTE: this does not have to necessarily involve Numba because it can also be
    #       the NumPy-based implementation under the hood
    func = _nb_chebyshev_polyvander if jit else _np_chebyshev_polyvander
    chebyshev_bases = func(  # type: ignore
        x=x_internal,  # type: ignore
        n=n,  # type: ignore
    )

    # --- Output post-processing ---

    # if only one kind is requested, the corresponding part of the output is returned
    # NOTE: the Array(s) has/have to be transposed because the low level functions
    #       return the transposed basis because it is more efficient for the computation
    if kind_internal in {1, 2}:
        return np.moveaxis(
            chebyshev_bases[kind_internal - 1],
            source=0,
            destination=1,
        )

    # if both kinds are requested, the full output is returned
    return tuple(  # type: ignore
        np.moveaxis(
            arr,
            source=0,
            destination=1,
        )
        for arr in chebyshev_bases
    )



"""
Module :mod:`_utils._x_preprocessing`

This module implements the input preprocessing for the independent variable ``x`` that
is shared by the Chebyshev polynomial and Hermite function classes, namely

- checking if ``x`` is still linked to the original data it was derived from and
- normalising ``x`` by shifting it with a center ``x_center`` and scaling it with a
    scaling factor ``alpha``

"""

# === Imports ===

from typing import Union

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ._validate import RealScalar

# === Functions ===


def _is_data_linked(
    arr: np.ndarray,
    original: Union[RealScalar, ArrayLike],
) -> bool:
    """
    Strictly checks for ``arr`` not sharing any data with ``original``, under the
    assumption that ``arr = atleast_1d(original)`` followed by a potential type
    conversion.
    If ``arr`` is a view of ``original``, this function returns ``False``.

    Was copied from the SciPy utility function ``scipy.linalg._misc._datacopied``, but
    the name and the docstring were adapted to make them clearer. Besides, the check for
    scalar ``original``s was added.

    """

    if np.isscalar(original):
        return False
    if arr is original:
        return True
    if not isinstance(original, np.ndarray) and hasattr(original, "__array__"):
        return original.__array__().dtype is arr.dtype  # type: ignore

    return arr.base is not None


def normalise_x_values(
    x_internal: NDArray[np.float64],
    x: Union[float, int, ArrayLike],
    x_center: float,
    alpha: float,
) -> NDArray[np.float64]:
    """
    Centers the given x-values around the given center value by handling potential
    copies and the special case where the center is the origin (0).
    Afterwards, the x-values are scaled with the scaling factor alpha, also handling
    the special case where alpha is 1.0.

    """

    # when the x-values are centered around the origin and not scaled, the x-values are
    # not modified at all
    center_is_origin = x_center == 0.0
    alpha_is_unity = alpha == 1.0
    if center_is_origin and alpha_is_unity:
        return x_internal

    # if x is a view of the original x-Array, a copy is made to avoid modifying
    # the x-values
    if _is_data_linked(arr=x_internal, original=x):
        x_internal = x_internal.copy()

    # the x-values are centered around the given center value (if required)
    if not center_is_origin:
        x_internal -= x_center

    # if required, the x-values are scaled with the scaling factor alpha (if required)
    if not alpha_is_unity:
        x_internal /= alpha

    return x_internal



"""
Module :mod:`chebyshev_polynomials._numba_funcs`

This module provides Numba-based implementations of the Chebyshev polynomials of the
first and second kind.

Depending on the runtime availability of Numba, the functions are either compiled or
imported from the NumPy-based implementation.

"""

# === Imports ===

from .._utils._numba_helpers import do_numba_normal_jit_action
from ._numpy_funcs import _chebyshev_polyvander

# === Functions ===

# NOTE: here are no functions because the NumPy-based implementation was written to be
#       compatible with Numba ``jit``-compilation


# === Compilation ===

# if available/enabled, the functions are compiled by Numba
try:
    if do_numba_normal_jit_action:  # pragma: no cover
        from numba import jit
    else:
        from .._utils import no_jit as jit

    # if it is enabled, the functions are compiled
    nb_chebyshev_polyvander = jit(
        nopython=True,
        cache=True,
    )(_chebyshev_polyvander)


# otherwise, the NumPy-based implementation of the Hermite functions is declared as the
# Numba-based implementation
except ImportError:  # pragma: no cover

    nb_chebyshev_polyvander = _chebyshev_polyvander



"""
Module :mod:`_utils`

This module provides utility functionalities that are used throughout the package, e.g.,

- handling of Numba-related tasks

"""

# === Imports ===

from ._numba_helpers import (  # noqa: F401
    NUMBA_NO_JIT_ARGV,
    NUMBA_NO_JIT_ENV_KEY,
    NumbaJitActions,
    do_numba_normal_jit_action,
    no_jit,
)
from ._validate import (  # noqa: F401
    IntScalar,
    RealScalar,
    get_validated_alpha,
    get_validated_chebpoly_or_hermfunc_input,
    get_validated_grid_points,
    get_validated_offset_along_axis,
    get_validated_order,
)
from ._x_preprocessing import normalise_x_values  # noqa: F401



"""
Package ``robust_fourier``

This package contains the implementation a noise- and outlier-robust Fourier transform
based on a Least Squares Fit of non-parametric functions to the data.

"""

# === Imports ===

import os as _os

from .chebyshev_polynomials import (  # noqa: F401
    ChebyshevPolynomialBasis,
    chebyshev_polyvander,
)
from .hermite_functions import (  # noqa: F401
    HermiteFunctionBasis,
    hermite_approx,
    hermite_function_vander,
    single_hermite_function,
)

# === Package Metadata ===

_AUTHOR_FILE_PATH = _os.path.join(_os.path.dirname(__file__), "AUTHORS.txt")
_VERSION_FILE_PATH = _os.path.join(_os.path.dirname(__file__), "VERSION.txt")

with open(_AUTHOR_FILE_PATH, "r") as author_file:
    __author__ = author_file.read().strip()

with open(_VERSION_FILE_PATH, "r") as version_file:
    __version__ = version_file.read().strip()



"""
Module :mod:`_utils.numba_helpers`

This module implements auxiliary functionalities to handle Numba-related tasks, such as

- checking whether Numba ``jit``-compilation has been explicitly specified to take no
    effect, e.g., for test coverage

"""

# === Imports ===

import os
from enum import Enum
from typing import Callable

# === Models ===

# an Enum that specifies the possible actions that can be taken regarding Numba
# ``jit``-compilation


class NumbaJitActions(Enum):
    """
    Specifies the possible actions that can be taken regarding Numba
    ``jit``-compilation.

    """

    NORMAL = "0"
    DEACTIVATE = "1"


# === Constants ===

# the runtime argument that is used to specify that Numba ``jit``-compilation should
# take no effect
NUMBA_NO_JIT_ARGV = "--no-jit"

# the environment variable that is used to specify that Numba ``jit``-compilation should
# take no effect
NUMBA_NO_JIT_ENV_KEY = "CUSTOM_NUMBA_NO_JIT"


# whether the environment variable is set to specify that Numba ``jit``-compilation
# should take effect or not in the current runtime environment
do_numba_normal_jit_action = (
    os.environ.get(NUMBA_NO_JIT_ENV_KEY, NumbaJitActions.NORMAL.value)
    == NumbaJitActions.NORMAL.value
)


# === Functions ===


def no_jit(*args, **kwargs) -> Callable:
    """
    Fake decorator that can be used to make sure that Numba ``jit``-compilation has no
    effect.

    Parameters
    ----------
    func : :class:`Callable`
        The function that is decorated.

    args : :class:`tuple`
        The fake positional arguments.

    kwargs : :class:`dict`
        The fake keyword arguments.

    Returns
    -------
    decorated_func : :class:`Callable`
        The decorated function.

    """

    def decorator(func: Callable) -> Callable:
        return func

    return decorator



"""
Module :mod:`hermite_functions._func_interface`

This module implements the interface to the either NumPy-based or Numba-based
implementations of the Hermite functions.

It augments them with an additional input validation which is better done in Python
and also handles the incorporation of the Numba-functions if Numba is available at
runtime.

"""

# === Imports ===

from math import sqrt as pysqrt
from typing import Optional, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray

from .._utils import (
    IntScalar,
    RealScalar,
    get_validated_chebpoly_or_hermfunc_input,
    get_validated_grid_points,
    normalise_x_values,
)
from ._numba_funcs import nb_hermite_function_vander as _nb_hermite_function_vander
from ._numpy_funcs import _hermite_function_vander as _np_hermite_function_vander
from ._numpy_funcs import _single_hermite_function as _np_single_hermite_function

# === Main Functions ===


def hermite_function_vander(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    jit: bool = True,
    validate_parameters: bool = True,
) -> NDArray[np.float64]:
    """
    Computes the complete basis (Vandermonde matrix) of dilated Hermite functions up to
    order ``n`` for the given points ``x``. It makes use of a recursion formula to
    compute all Hermite basis functions in one go.

    Parameters
    ----------
    x : :class:`float` or :class:`int` or Array-like of shape (m,)
        The points at which the dilated Hermite functions are evaluated.
        Internally, it will be promoted to ``np.float64``.
        It has to contain at least one element.
    n : :class:`int`
        The order of the dilated Hermite functions.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite functions.
        If ``None`` or ``0``, the functions are centered at the origin.
        Otherwise, the center is shifted to the given value.
    jit : :class:`bool`, default=``True``
        Whether to use the Numba-accelerated implementation (``True``) or the
        NumPy-based implementation (``False``).
        If Numba is not available, the function silently falls back to the NumPy-based
        implementation.
    validate_parameters : :class:`bool`, default=``True``
        Whether to validate all the input parameters (``True``) or only ``x``
        (``False``).
        Disabling the input checks is not recommended and was only implemented for
        internal use.

    Returns
    -------
    hermite_func_vander : :class:`numpy.ndarray` of shape (m, n + 1) of dtype ``np.float64``
        The values of the dilated Hermite functions at the points ``x`` represented as
        a Vandermonde matrix.
        It will always be 2D even if ``x`` is a scalar.

    Raises
    ------
    TypeError
        If any of the input arguments is not of the expected type.
    ValueError
        If ``x`` is not 1-dimensional after conversion to a NumPy array.
    ValueError
        If ``n`` is not a non-negative integer or ``alpha`` is not a positive number.

    Notes
    -----
    The dilated Hermite functions are defined as

    .. image:: docs/hermite_functions/equations/HF-01-Hermite_Functions_TimeSpace_Domain.svg

    with the Hermite polynomials

    .. image:: docs/hermite_functions/equations/HF-02-Hermite_Polynomials_TimeSpace_Domain.svg

    Internally, they are computed in a numerically stable way that relies on a
    logarithmic scaling trick to avoid over- and underflow in the recursive calculation
    of the Hermite polynomials and the Gaussians. This allows for arbitrary large orders
    ``n`` to be evaluated.

    References
    ----------
    The implementation is an adaption of the Appendix in [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    """  # noqa: E501

    # --- Input validation ---

    if validate_parameters:
        (
            x_internal,
            n,
            alpha,
            x_center,
        ) = get_validated_chebpoly_or_hermfunc_input(
            x=x,
            x_dtype=np.float64,
            n=n,
            alpha=alpha,
            x_center=x_center,
        )

    else:  # pragma: no cover
        x_internal = get_validated_grid_points(grid_points=x, dtype=np.float64)

    # --- Computation ---

    # if required, the x-values are centered
    x_internal = normalise_x_values(
        x_internal=x_internal,
        x=x,
        x_center=x_center,  # type: ignore
        alpha=alpha,  # type: ignore
    )

    # if requested, the Numba-accelerated implementation is used
    # NOTE: this does not have to necessarily involve Numba because it can also be
    #       the NumPy-based implementation under the hood
    func = _nb_hermite_function_vander if jit else _np_hermite_function_vander
    hermite_basis = func(  # type: ignore
        x=x_internal,  # type: ignore
        n=n,  # type: ignore
    )

    # to preserve orthonormality, the Hermite functions are scaled with the square root
    # of the scaling factor alpha (if required)
    if alpha != 1.0:
        hermite_basis *= 1.0 / pysqrt(alpha)

    # NOTE: the Array has to be transposed because the low level functions return the
    #       transposed basis because it is more efficient for the computation
    return np.moveaxis(
        hermite_basis,
        source=0,
        destination=1,
    )


def single_hermite_function(
    x: Union[RealScalar, ArrayLike],
    n: IntScalar,
    alpha: RealScalar = 1.0,
    x_center: Optional[RealScalar] = None,
    validate_parameters: bool = True,
) -> NDArray[np.float64]:
    """
    Computes a single dilated Hermite function of order ``n`` for the given points
    ``x``. It offers a fast alternative for the computation of only a single high order
    Hermite function (``n >= 1000``), but not a full basis of them.

    Parameters
    ----------
    x : :class:`float` or :class:`int` or Array-like of shape (m,)
        The points at which the dilated Hermite function is evaluated.
        Internally, it will be promoted to ``np.float64``.
        It has to contain at least one element.
    n : :class:`int`
        The order of the dilated Hermite function.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite function.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.
    validate_parameters : :class:`bool`, default=``True``
        Whether to validate all the input parameters (``True``) or only ``x``
        (``False``).
        Disabling the input checks is not recommended and was only implemented for
        internal use.

    Returns
    -------
    hermite_function : :class:`numpy.ndarray` of shape (m,) of dtype ``np.float64``
        The values of the dilated Hermite function at the points ``x``.
        It will always be 1D even if ``x`` is a scalar.

    Raises
    ------
    TypeError
        If any of the input arguments is not of the expected type.
    ValueError
        If ``x`` is not 1-dimensional after conversion to a NumPy array.
    ValueError
        If ``n`` is not a non-negative integer or ``alpha`` is not a positive number.

    Notes
    -----
    The dilated Hermite functions are defined as

    .. image:: docs/hermite_functions/equations/HF-01-Hermite_Functions_TimeSpace_Domain.svg

    with the Hermite polynomials

    .. image:: docs/hermite_functions/equations/HF-02-Hermite_Polynomials_TimeSpace_Domain.svg

    For their computation, the function does not rely on recursion, but a direct
    evaluation of the Hermite functions via a complex integral.
    While this may be way faster than using the recursion for a single function for
    ``n >= 1000``, it is not suitable for computing a full basis of them.

    References
    ----------
    The implementation is based on [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    """  # noqa: E501

    # --- Input validation ---

    if validate_parameters:
        (
            x_internal,
            n,
            alpha,
            x_center,
        ) = get_validated_chebpoly_or_hermfunc_input(
            x=x,
            x_dtype=np.float64,
            n=n,
            alpha=alpha,
            x_center=x_center,
        )

    else:  # pragma: no cover
        x_internal = get_validated_grid_points(grid_points=x, dtype=np.float64)

    # --- Computation ---

    # if required, the x-values are centered
    x_internal = normalise_x_values(
        x_internal=x_internal,
        x=x,
        x_center=x_center,  # type: ignore
        alpha=alpha,  # type: ignore
    )

    hermite_function = _np_single_hermite_function(
        x=x_internal,
        n=n,  # type: ignore
    )

    # to preserve orthonormality, the Hermite function is scaled with the square root of
    # the scaling factor alpha (if required)
    if alpha != 1.0:
        hermite_function *= 1.0 / pysqrt(alpha)

    return hermite_function



"""
Module :mod:`hermite_functions._class_interface`

This module implements a class based interface to the Hermite functions via the class
:class:`HermiteFunctionBasis`.

"""

# === Imports ===

from typing import Optional, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray

from .._utils import (
    IntScalar,
    RealScalar,
    get_validated_alpha,
    get_validated_offset_along_axis,
    get_validated_order,
)
from ._func_interface import hermite_function_vander

# === Classes ===


class HermiteFunctionBasis:
    """
    This class represents a basis of Hermite functions which are defined by their

    - order ``n``,
    - scaling factor ``alpha``, and
    - center ``x_center`` (mu)

    For their definition, please refer to the Notes section.

    Parameters
    ----------
    n : :class:`int`
        The order of the dilated Hermite functions.
        It must be a non-negative integer ``>= 0``.
    alpha : :class:`float` or :class:`int`, default=``1.0``
        The scaling factor of the independent variable ``x`` for
        ``x_scaled = x / alpha``.
        It must be a positive number ``> 0``.
    x_center : :class:`float` or :class:`int` or ``None``, default=``None``
        The center of the dilated Hermite functions.
        If ``None`` or ``0``, the function is centered at the origin.
        Otherwise, the center is shifted to the given value.
    jit : :class:`bool`, default=``True``
        Whether to use the Numba-accelerated implementation (``True``) or the
        NumPy-based implementation (``False``).
        If Numba is not available, the function silently falls back to the NumPy-based
        implementation.

    Attributes
    ----------
    n : :class:`int`
        The order of the Hermite functions.
    alpha : :class:`float`
        The scaling factor of the independent variables.
    x_center : :class:`float`
        The x-center mu of the Hermite functions.

    Methods
    -------
    __len__()
        Returns the number of Hermite basis functions that will be computed with the
        given parameters.
    eval(x, n, alpha, x_center, validate_parameters=True)
        Evaluates the Hermite functions at the given points with the specified
        parameters.
    __call__(x)
        Evaluates the Hermite functions at the given points with the class parameters.

    Raises
    ------
    TypeError
        If any of the input parameters is not of the expected type.
    ValueError
        If ``n`` is not a non-negative integer or ``alpha`` is not a positive number.

    Notes
    -----
    The dilated Hermite functions are defined as

    .. image:: docs/hermite_functions/equations/HF-01-Hermite_Functions_TimeSpace_Domain.svg

    with the Hermite polynomials

    .. image:: docs/hermite_functions/equations/HF-02-Hermite_Polynomials_TimeSpace_Domain.svg

    Internally, they are computed in a numerically stable way that relies on a
    logarithmic scaling trick to avoid over- and underflow in the recursive calculation
    of the Hermite polynomials and the Gaussians. This allows for arbitrary large orders
    ``n`` to be evaluated.

    References
    ----------
    The implementation is an adaption of the Appendix in [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    """  # noqa: E501

    # --- Constructor ---

    def __init__(
        self,
        n: IntScalar,
        alpha: RealScalar = 1.0,
        x_center: Optional[RealScalar] = None,
        jit: bool = True,
    ):
        self._n: int = get_validated_order(n=n)
        self._alpha: float = get_validated_alpha(alpha=alpha)
        self._x_center: float = get_validated_offset_along_axis(
            offset=x_center,
            which_axis="x",
        )
        self._jit: bool = jit

    # --- Properties ---

    @property
    def n(self) -> int:
        return self._n

    @n.setter
    def n(self, value: IntScalar) -> None:
        self._n = get_validated_order(n=value)

    @property
    def alpha(self) -> float:
        return self._alpha

    @alpha.setter
    def alpha(self, value: RealScalar) -> None:
        self._alpha = get_validated_alpha(alpha=value)

    @property
    def x_center(self) -> float:
        return self._x_center

    @x_center.setter
    def x_center(self, value: Optional[RealScalar]) -> None:
        self._x_center = get_validated_offset_along_axis(
            offset=value,
            which_axis="x",
        )

    @property
    def jit(self) -> bool:
        return self._jit

    @jit.setter
    def jit(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(
                f"Expected 'jit' to be a boolean but it is of type "
                f"'{type(value).__name__}'."
            )

        self._jit = value

    # --- Public Methods ---

    @staticmethod
    def eval(
        x: Union[RealScalar, ArrayLike],
        n: IntScalar = 10,
        alpha: RealScalar = 1.0,
        x_center: Optional[RealScalar] = None,
        jit: bool = True,
        validate_parameters: bool = True,
    ) -> NDArray[np.float64]:
        """
        Evaluates the Hermite functions at the given points with the specified
        parameters.

        Parameters
        ----------
        x : :class:`float` or :class:`int` or Array-like of shape (m,)
            The points at which the dilated Hermite functions are evaluated.
            Internally, it will be promoted to ``np.float64``.
            It has to contain at least one element.
        n : :class:`int`
            The order of the dilated Hermite functions.
            It must be a non-negative integer ``>= 0``.
        alpha : :class:`float` or :class:`int`, default=``1.0``
            The scaling factor of the independent variable ``x`` for
            ``x_scaled = x / alpha``.
            It must be a positive number ``> 0``.
        x_center : :class:`float` or :class:`int` or ``None``, default=``None``
            The center of the dilated Hermite functions.
            If ``None`` or ``0``, the functions are centered at the origin.
            Otherwise, the center is shifted to the given value.
        jit : :class:`bool`, default=``True``
            Whether to use the Numba-accelerated implementation (``True``) or the
            NumPy-based implementation (``False``).
            If Numba is not available, the function silently falls back to the
            NumPy-based implementation.
        validate_parameters : :class:`bool`, default=``True``
            Whether to validate ``n``, ``alpha``, and ``x_center`` before evaluating the
            Hermite functions.
            Disabling the checks is highly discouraged and was only implemented for
            internal purposes.

        Returns
        -------
        hermite_basis : :class:`numpy.ndarray` of shape (m, n + 1) of dtype ``np.float64``
            The values of the dilated Hermite functions at the points ``x``.
            It will always be 2D even if ``x`` is a scalar.

        Notes
        -----
        For a detailed description of the Hermite functions, please refer to the
        docstring of the class :class:`HermiteFunctionBasis` itself.

        """  # noqa: E501

        return hermite_function_vander(
            x=x,
            n=n,
            alpha=alpha,
            x_center=x_center,
            jit=jit,
            validate_parameters=validate_parameters,
        )

    # --- Magic Methods ---

    def __len__(self):
        """
        Returns the number of Hermite basis functions that will be computed, i.e.,
        ``n + 1``.

        """

        return self._n + 1

    def __call__(
        self,
        x: Union[RealScalar, ArrayLike],
    ) -> NDArray[np.float64]:
        """
        Evaluates the Hermite functions at the given points with the class parameters.
        Please refer to the docstrings of the class itself and the method :meth:`eval`
        for more information.

        """  # noqa: E501

        return self.eval(
            x=x,
            n=self._n,
            alpha=self._alpha,
            x_center=self._x_center,
            jit=self._jit,
            validate_parameters=False,
        )



"""
Module :mod:`fourier_transform._fft_utils`

This module provides utility functions for the Fourier transform, such as

- the computation of the angular frequency grid for the discrete Fourier transform
- the conversion from a discrete Fourier transform to a continuous Fourier transform
    and vice versa

"""

# === Imports ===

from dataclasses import dataclass, field
from math import sqrt as pysqrt
from typing import Literal, Union

import numpy as np

# === Constants ===

NORM_BACKWARD = "backward"
NORM_ORTHOGONAL = "ortho"
NORM_FORWARD = "forward"

# === Types ===

_DftNorms = Literal["backward", "ortho", "forward"]
_FTConversionsTo = Literal["continuous", "discrete"]

# === Models ===


@dataclass
class TimeSpaceSignal:
    """
    Dataclass to represent the grid points and the signal in the time or space domain.

    Attributes
    ----------
    y : :class:`numpy.ndarray` of shape ``(n,)``
        The signal in the time or space domain.
    x : :class:`numpy.ndarray` of shape ``(n,)``
        The grid points in the time or space domain.
        It has to be evenly spaced and sorted in ascending order.
        If empty, an index-based grid is assumed.
    delta_x : :class:`float`
        The grid spacing.
        It is computed from the grid points.

    Notes
    -----
    The number of grid points can be obtained by ``len(time_space_domain)``.

    """

    y: np.ndarray
    x: np.ndarray = field(default_factory=lambda: np.array([]))

    delta_x: float = field(init=False)

    def __post_init__(self) -> None:
        # if there are no grid points, an index-based grid is assumed
        if self.x.size < 1:
            self.x = np.arange(
                start=0,
                stop=self.y.size,
                step=1,
                dtype=np.int64,
            ).astype(np.float64)

        # the grid spacing is computed from the grid points
        self.delta_x = grid_spacing(x=self.x)

    def __len__(self) -> int:
        return self.y.size


@dataclass
class _BaseFourierTransform:
    """
    Base class for the representation of a Fourier transform.

    Attributes
    ----------
    angular_frequencies : :class:`numpy.ndarray` of shape ``(n,)``
        The angular frequency grid points.
    time_space_signal : :class:`TimeSpaceDomainRepresentation`
        The grid points and the signal in the time or space domain.

    """

    angular_frequencies: np.ndarray
    time_space_signal: TimeSpaceSignal


@dataclass
class DiscreteFourierTransform(_BaseFourierTransform):
    """
    Dataclass to represent a discrete Fourier transform.

    Attributes
    ----------
    angular_frequencies : :class:`numpy.ndarray` of shape ``(n,)``
        The angular frequency grid points.
    dft : :class:`numpy.ndarray` of shape ``(n,)``
        The discrete Fourier transform.
    norm : {"backward", "ortho", "forward"}
        The normalization to use.
        Please refer to :func:`discrete_ft` for more details.
    time_space_signal : :class:`TimeSpaceDomainRepresentation`
        The grid points and the signal in the time or space domain.

    Properties
    ----------
    real : :class:`numpy.ndarray` of shape ``(n,)``
        The real part of the discrete Fourier transform.
    imag : :class:`numpy.ndarray` of shape ``(n,)``
        The imaginary part of the discrete Fourier transform.

    """

    dft: np.ndarray
    norm: _DftNorms

    @property
    def real(self) -> np.ndarray:
        return self.dft.real

    @property
    def imag(self) -> np.ndarray:
        return self.dft.imag


@dataclass
class ContinuousFourierTransform(_BaseFourierTransform):
    """
    Dataclass to represent a continuous Fourier transform.

    Attributes
    ----------
    angular_frequencies : :class:`numpy.ndarray` of shape ``(n,)``
        The angular frequency grid points.
    cft : :class:`numpy.ndarray` of shape ``(n,)``
        The continuous Fourier transform.
    time_space_signal : :class:`TimeSpaceDomainRepresentation`
        The grid points and the signal in the time or space domain.

    Properties
    ----------
    real : :class:`numpy.ndarray` of shape ``(n,)``
        The real part of the continuous Fourier transform.
    imag : :class:`numpy.ndarray` of shape ``(n,)``
        The imaginary part of the continuous Fourier transform.

    """

    cft: np.ndarray

    @property
    def real(self) -> np.ndarray:
        return self.cft.real

    @property
    def imag(self) -> np.ndarray:
        return self.cft.imag


# === Functions ===


def grid_spacing(x: np.ndarray) -> float:
    """
    Computes the grid spacing from a an evenly spaced grid.

    Parameters
    ----------
    x : :class:`numpy.ndarray` of shape ``(n,)``
        The grid points.
        It has to be evenly spaced and sorted in strictly ascending order. Strictly
        ascending means that the difference between two consecutive grid points is
        strictly positive.

    Returns
    -------
    d : :class:`float`
        The grid spacing.

    Raises
    ------
    ValueError
        If the grid is not evenly spaced.

    """

    delta_x = (x[-1] - x[0]) / (x.size - 1)
    diff_x = np.diff(x)
    if (diff_x <= 0.0).any():
        raise ValueError("The grid points are not sorted in strictly ascending order.")

    if not np.allclose(
        diff_x,
        np.full(shape=(x.size - 1,), fill_value=delta_x),
    ):
        raise ValueError("The grid points are not evenly spaced.")

    return delta_x


def angular_frequency_grid(
    n: int,
    d: Union[float, int] = 1.0,
) -> np.ndarray:
    """
    Compute the angular frequency grid for the discrete Fourier transform.
    It is given by :math:`\\angular_frequencies = \\frac{2\\cdot\\pi\\cdot k}{n\\cdot d}`
    where

    * :math:`n` is the number of grid points
    * :math:`d` is the grid spacing

    Parameters
    ----------
    n : :class:`int`
        The number of grid points.
    d : :class:`float` or :class:`int`, default=1.0
        The grid spacing.

    Returns
    -------
    angular_frequencies : :class:`numpy.ndarray` of shape ``(n,)``
        The angular frequency grid.

    """  # noqa: E501

    return 2 * np.pi * np.fft.fftfreq(n=n, d=d)


def discrete_ft(
    signal: TimeSpaceSignal,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> DiscreteFourierTransform:
    """
    Object-oriented wrapper for the NumPy function :func:`numpy.fft.fft` to compute the
    discrete Fourier transform of a signal.

    Parameters
    ----------
    signal : :class:`TimeSpaceSignal`
        The signal in the time or space domain.
        For more details, please refer to :class:`TimeSpaceSignal`.
    norm : {"backward", "ortho", "forward"}, default="backward"
        The normalization to use.

        * ``"backward"``: no normalization is applied. Then, the inverse Fourier
            transform has to be multiplied by ``1/n``.
        * ``"ortho"``: the normalization factor is ``1/sqrt(n)`` for both the forward
            and backward Fourier transform.
        * ``"forward"``: the normalization factor is ``1/n`` for the Fourier transform
            and no normalization is needed for the inverse Fourier transform.

    Returns
    -------
    discrete_ft : :class:`DiscreteFourierTransform`
        The discrete Fourier transform.
        For more details, please refer to :class:`DiscreteFourierTransform`.

    """

    # the discrete Fourier transform is computed together with the angular frequency
    # grid ...
    angular_frequencies = angular_frequency_grid(n=len(signal), d=signal.delta_x)
    discrete_ft = np.fft.fft(a=signal.y, norm=norm)

    # ... and stored in a dataclass
    return DiscreteFourierTransform(
        angular_frequencies=angular_frequencies,
        dft=discrete_ft,
        norm=norm,
        time_space_signal=signal,
    )


def _ft_conversion_factors(
    x: np.ndarray,
    delta_x: float,
    angular_frequencies: np.ndarray,
    norm: _DftNorms,
    to_kind: _FTConversionsTo,
) -> np.ndarray:
    """
    Compute the factors for the conversion from a discrete Fourier transform to a
    continuous Fourier transform or vice versa.

    For the conversion from a discrete Fourier transform to a continuous Fourier
    transform, the correction is given by

    .. code-block:: python

        correction_factors = (phi * delta_x / np.sqrt(2 * np.pi)) *  * np.exp(-1j * angular_frequencies * x[0])
        continuous_ft = discrete_ft * correction_factors

    while for the conversion from a continuous Fourier transform to a discrete Fourier
    transform, the correction is given by

    .. code-block:: python

        correction_factors = (np.sqrt(2 * np.pi) / phi / delta_x) * np.exp(1j * angular_frequencies * x[0])
        discrete_ft = continuous_ft * correction_factors

    Depending on the normalization, the factor :math:`\\phi` is different:

    * ``"backward"``: :math:`\\phi = 1`
    * ``"ortho"``: :math:`\\phi = \\sqrt{n}`
    * ``"forward"``: :math:`\\phi = n`

    where :math:`n` is the number of grid points.

    Parameters
    ----------
    x : :class:`numpy.ndarray` of shape ``(n,)``
        The grid points in the time or space domain.
    delta_x : :class:`float`
        The grid spacing.
    angular_frequencies : :class:`numpy.ndarray` of shape ``(n,)``
        The angular frequency grid.
    norm : {"backward", "ortho", "forward"}, default="backward"
        The normalization to use.
        Please refer to :func:`discrete_ft` for more details.
    from_kind : {"continuous", "discrete"}, default="discrete"
        Whether the conversion is from a discrete Fourier transform to a continuous
        Fourier transform (``"continuous"``) or vice versa (``"discrete"``).

    Returns
    -------
    correction_factors : :class:`numpy.ndarray` of shape ``(n,)``
        The correction factors.

    Raises
    ------
    ValueError
        If the normalization is unknown.

    """  # noqa: E501

    norm_inter = norm.lower()
    if norm_inter == NORM_BACKWARD:
        phi = 1.0
    elif norm_inter == NORM_ORTHOGONAL:
        phi = pysqrt(x.size)
    elif norm_inter == NORM_FORWARD:
        phi = x.size
    else:  # pragma: no cover
        raise ValueError(f"Unknown normalization '{norm}'.")

    if to_kind == "continuous":
        return (phi * delta_x / pysqrt(2.0 * np.pi)) * np.exp(
            -1j * angular_frequencies * x[0]
        )
    else:
        return (pysqrt(2.0 * np.pi) / phi / delta_x) * np.exp(
            1j * angular_frequencies * x[0]
        )


def convert_discrete_to_continuous_ft(
    dft: DiscreteFourierTransform,
) -> ContinuousFourierTransform:
    """
    Convert a discrete Fourier transform to a continuous Fourier transform.

    Parameters
    ----------
    dft : :class:`DiscreteFourierTransform`
        The discrete Fourier transform.
        For more details, please refer to :class:`DiscreteFourierTransform`.

    Returns
    -------
    cft : :class:`ContinuousFourierTransform`
        The continuous Fourier transform.
        For more details, please refer to :class:`ContinuousFourierTransform`.

    """

    # the conversion factors are computed ...
    conversion_factors = _ft_conversion_factors(
        x=dft.time_space_signal.x,
        delta_x=dft.time_space_signal.delta_x,
        angular_frequencies=dft.angular_frequencies,
        norm=dft.norm,
        to_kind="continuous",
    )

    # ... and the continuous Fourier transform is computed
    return ContinuousFourierTransform(
        angular_frequencies=dft.angular_frequencies,
        cft=dft.dft * conversion_factors,
        time_space_signal=dft.time_space_signal,
    )


def convert_continuous_to_discrete_ft(
    cft: ContinuousFourierTransform,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> DiscreteFourierTransform:
    """
    Convert a continuous Fourier transform to a discrete Fourier transform.

    Parameters
    ----------
    cft : :class:`ContinuousFourierTransform`
        The continuous Fourier transform.
        For more details, please refer to :class:`ContinuousFourierTransform`.
    norm : {"backward", "ortho", "forward"}, default="backward"
        The normalization to use.
        Please refer to :func:`discrete_ft` for more details.

    Returns
    -------
    dft : :class:`DiscreteFourierTransform`
        The discrete Fourier transform.
        For more details, please refer to :class:`DiscreteFourierTransform`.

    """

    # the conversion factors are computed ...
    conversion_factors = _ft_conversion_factors(
        x=cft.time_space_signal.x,
        delta_x=cft.time_space_signal.delta_x,
        angular_frequencies=cft.angular_frequencies,
        norm=norm,
        to_kind="discrete",
    )

    # ... and the discrete Fourier transform is computed
    return DiscreteFourierTransform(
        angular_frequencies=cft.angular_frequencies,
        dft=cft.cft * conversion_factors,
        norm=norm,
        time_space_signal=cft.time_space_signal,
    )



"""
Module :mod:`chebyshev_polynomials._numpy_funcs`

This module provides NumPy-based implementations of the Chebyshev polynomials of the
first and second kind.

"""

# === Imports ===

from typing import Tuple

import numpy as np

# === Functions ===


def _chebyshev_polyvander(
    x: np.ndarray,
    n: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Simultaneously evaluates the complete basis (Vandermonde matrix) of Chebyshev
    polynomials of the first and second kind by making use of a numerically stabilised
    combined recurrence relation.

    Parameters
    ----------
    x : :class:`numpy.ndarray` of shape (m,)
        The points at which the Chebyshev polynomials are evaluated.
    n : :class:`int`
        The order of the Chebyshev polynomials.

    Returns
    -------
    chebyshev_t1_vander : :class:`numpy.ndarray` of shape (m, n + 1)
        The values of the Chebyshev polynomials of the first kind evaluated at the
        points ``x`` represented as a Vandermonde matrix.

    chebyshev_u2_vander : :class:`numpy.ndarray` of shape (m, n + 1)
        The values of the Chebyshev polynomials of the second kind evaluated at the
        points ``x`` represented as a Vandermonde matrix.

    References
    ----------
    The implementation is taken from [1]_.

    .. [1] Hrycak T., Schmutzhard S., Accurate evaluation of Chebyshev polynomials in
       floating-point arithmetic, BIT Numer Math, (2019) 59, pp. 403–416,
       DOI: 10.1007/s10543-018-0738-5

    Notes
    -----
    The naive three-term recurrence relations for Chebyshev polynomials of the first and
    second kind have bad numerical properties because the error is increasing
    proportional to ``n * n``, i.e., quadratically with the order of the polynomial
    ``n``.
    However, a combined recurrence relation that links both kinds of Chebyshev
    polynomials has an error proportional to ``n``, i.e., a linear increase with the
    order of the polynomial ``n``. Already for ``n`` as little as ``4``, the combined
    recurrence relation outperforms the naive three-term recurrence relations. So, the
    roughly doubled computational cost of the combined recurrence relation is justified.
    On top of that, a combined evaluation offers the opportunity to easily compute the
    derivatives of the Chebyshev polynomials because the derivatives of the first kind
    are related to the polynomials of the second kind and vice versa.

    """

    # some pre-computations are done to avoid redundant computations
    # the combined recurrence relation is started with the initial value
    # - 1 for the Chebyshev polynomial of the first kind T_0(x)
    # - 0 for the Chebyshev polynomial of the second kind U_{-1}(x)
    chebyshev_t1_vander = np.empty(shape=(n + 1, x.size))
    chebyshev_u2_vander = np.empty_like(chebyshev_t1_vander)
    t_i_minus_1 = np.ones_like(x)
    u_i_minus_2 = np.zeros_like(x)
    one_minus_x_squared = 1.0 - (x * x)

    # then, the combined recurrence relation is iteratively applied by means of a loop
    # the recurrence relation is given by
    # T_{n}(x) = x * T_{n-1}(x) - (1 - x * x) * U_{n-2}(x)
    # U_{n-1}(x) = x * U_{n-2}(x) + T_{n-1}(x)
    for iter_j in range(0, n + 1):
        chebyshev_t1_vander[iter_j] = t_i_minus_1  # NOTE: is not a view
        u_i_minus_1 = x * u_i_minus_2 + t_i_minus_1
        chebyshev_u2_vander[iter_j] = u_i_minus_1  # NOTE: is not a view
        t_i_minus_1 = x * t_i_minus_1 - one_minus_x_squared * u_i_minus_2
        u_i_minus_2 = u_i_minus_1

    return chebyshev_t1_vander, chebyshev_u2_vander



"""
Module :mod:`chebyshev_polynomials`

This module provides implementations of the Chebyshev polynomials of the first and
second kind for fitting a basis for robust Fourier transforms.

"""

# === Imports ===

from ._class_interface import ChebyshevPolynomialBasis  # noqa: F401
from ._func_interface import chebyshev_polyvander  # noqa: F401



"""
Module :mod:`fourier_transform`

This module provides the implementation of the robust Fourier transform by Least Squares
Fits with Hermite functions as well as general utility functions for dealing with
Discrete Fourier transforms.

"""

# === Imports ===

from ._fft_utils import (  # noqa: F401
    ContinuousFourierTransform,
    DiscreteFourierTransform,
    TimeSpaceSignal,
    angular_frequency_grid,
    convert_continuous_to_discrete_ft,
    convert_discrete_to_continuous_ft,
    discrete_ft,
    grid_spacing,
)



"""
Module :mod:`hermite_functions._numpy_funcs`

This module provides NumPy-based implementations of the Hermite functions.

"""

# === Imports ===

from math import ceil as py_ceil
from math import sqrt as py_sqrt

import numpy as np
from numpy import abs as np_abs
from numpy import exp, log, sqrt, square
from scipy.integrate import trapezoid

# === Functions ===


def _hermite_function_vander(
    x: np.ndarray,
    n: int,
) -> np.ndarray:
    """
    Evaluates the complete basis (Vandermonde matrix) of Hermite functions that are
    given by the product of a scaled Gaussian with a Hermite polynomial and can be
    written as follows:

    .. image:: docs/hermite_functions/equations/HF-05-Hermite_Functions_Basic_Definition.svg

    .. image:: docs/hermite_functions/equations/HF-06-Hermite_Polynomials_Basic_Definition.svg

    Please refer to the Notes section for further details.

    Parameters
    ----------
    x : :class:`numpy.ndarray` of shape (m,)
        The points at which the Hermite functions are evaluated.
    n : :class:`int`
        The order of the Hermite functions.

    Returns
    -------
    hermite_func_vander : :class:`numpy.ndarray` of shape (m, n + 1)
        The values of the Hermite functions evaluated at the points ``x`` represented as
        a Vandermonde matrix.

    References
    ----------
    The implementation is an adaption of the Appendix in [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    Notes
    -----
    Direct evaluation of the Hermite polynomials becomes ill-conditioned in finite
    precision since it involves the division of a polynomial by the square root of
    a factorial of ``n`` which is prone to overflow.
    To avoid this, the recurrence relation of the Hermite functions is used to compute
    the values in a numerically stable way.
    However, there will still be underflow of the Gaussian part of the Hermite function
    for large values of ``x``. Therefore, a special scaling strategy is employed that
    keeps the values of the Hermite functions at a scale of 1 during the recursion while
    tracking a correction term that is added to the exponent of the Gaussian part of the
    Hermite function.

    """  # noqa: E501

    # the recurrence relation here is started from the virtual -1-th order Hermite
    # function which is defined as h_{-1} = 0
    # this is done to make the recursion easier and to avoid the need for handling
    # too many special cases
    h_i_minus_1 = np.zeros_like(x)

    # a result Array for the results is initialised
    hermite_func_vander = np.empty(shape=(n + 1, x.size))

    # the 0-th order Hermite function is defined as
    # h_{0} = pi ** (-1/4) * exp(-x ** 2 / 2)
    # NOTE: here it is kept as h_{0} = exp(phi) * 1 where phi is the exponent
    #       "correction" term phi = (ln(pi) / 4) - 0.5 * (x ** 2)
    log_fourth_root_of_pi = -0.28618247146235  # ln(pi) / 4
    h_i = np.ones_like(x)
    exponent_corrections = log_fourth_root_of_pi - 0.5 * square(x)

    hermite_func_vander[0] = exp(exponent_corrections)

    # if only the 0-th order is requested, the function can exit early here
    if n < 1:
        return hermite_func_vander

    # if higher orders are requested, a recursion is entered to compute the remaining
    # Hermite functions
    # the recursion is given by
    # h_{i+1} = sqrt(2 / (i + 1)) * x * h_{i} - sqrt(i / (i + 1)) * h_{i-1}
    # this is done in a numerically stable way by keeping the Hermite functions at a
    # scale of one and keeping track of the updated correction factors phi

    # the pre-factors for h_{i} and h_{i-1} are pre-computed
    # the pre-factor for h_{i} is sqrt(2 / (i + 1)) * x (here without the x-part)
    iterators = np.arange(0, n, 1, np.int64)
    prefactors_i = sqrt(2.0 / (iterators + 1.0))
    # the pre-factor for h_{i-1} is sqrt(i / (i + 1))
    prefactors_i_minus_1 = sqrt(iterators / (iterators + 1.0))

    for iter_i in iterators:
        # the new Hermite function is computed ...
        h_i_plus_1 = (
            prefactors_i[iter_i] * x * h_i - prefactors_i_minus_1[iter_i] * h_i_minus_1
        )
        # ... and stored after the correction factor is applied
        hermite_func_vander[iter_i + 1] = exp(exponent_corrections) * h_i_plus_1

        # afterwards, the correction factors are updated
        # NOTE: special care must be taken for values that are zero to avoid division by
        #       zero; they will not be updated
        scale_factors = np.where(h_i_plus_1 != 0.0, np_abs(h_i_plus_1), 1.0)
        h_i_minus_1 = h_i / scale_factors
        # NOTE: theoretically, h_{i+1} would be divided by its absolute value here, but
        #       a / |a| = sign(a) so the expensive division can be stated as a sign
        #       evaluation; here, everything relies on a sign definition that gives 0
        #       for a value of 0 and not +1 or -1 and ``np.sign`` meets this requirement
        h_i = np.sign(h_i_plus_1)
        exponent_corrections += log(scale_factors)

    # finally, the Hermite functions are returned
    return hermite_func_vander


def _single_hermite_function(
    x: np.ndarray,
    n: int,
) -> np.ndarray:
    """
    Computes a single Hermite function of order ``n`` at the given points ``x`` by
    making use of a complex integral which is way faster than the recursion if only
    a specific Hermite function is needed.

    The Hermite functions are defined as

    .. image:: docs/hermite_functions/equations/HF-05-Hermite_Functions_Basic_Definition.svg

    .. image:: docs/hermite_functions/equations/HF-06-Hermite_Polynomials_Basic_Definition.svg

    Parameters
    ----------
    x : :class:`numpy.ndarray` of shape (m,)
        The points at which the Hermite functions are evaluated.
    n : :class:`int`
        The order of the Hermite functions.

    Returns
    -------
    hermite_function : :class:`numpy.ndarray` of shape (m,)
        The values of the ``n``-th order Hermite function.

    References
    ----------
    The implementation is based on [1]_.

    .. [1] Bunck B. F., A fast algorithm for evaluation of normalized Hermite
       functions, BIT Numer Math (2009), 49, pp. 281–295, DOI: 10.1007/s10543-009-0216-1

    """  # noqa: E501

    # for the special case of n = 0, the Hermite function is just the Gaussian
    if n == 0:
        log_fourth_root_of_pi = -0.28618247146235  # ln(pi) / 4
        return np.exp(log_fourth_root_of_pi - 0.5 * np.square(x))

    # the integration of the circular integral from -pi to pi is set up
    # the number of required data points to evaluate this integral properly is
    # determined by the order as ``num_integ_points = ceil(12.42 * sqrt(n))``
    # NOTE: here, a bit more than half of this number is chosen because the function
    #       is symmetric around the y-axis
    # NOTE: the second formula comes from a custom regression with the aim to keep the
    #       error bounded to the machine precision for all the small orders where the
    #       approximation made in the publication is invalid; it will naturally fade
    #       out for larger orders where the approximation is valid; the turnover point
    #       is around n = 150
    sqrt_n = py_sqrt(n)
    num_integ_points = max(
        py_ceil(6.5 * sqrt_n),
        py_ceil(-0.173858887140972 * n + 8.03260560462478 * sqrt_n + 6.85116548584779),
    )
    # the points are evenly distributed over the positive half of the circle
    # NOTE: we will exploit mirroring later
    integ_points = np.linspace(0.0, np.pi, num_integ_points)
    delta_angle = np.pi / (num_integ_points - 1)

    # then, a correction factor gamma needs to be evaluated from a value k which is
    # sqrt(n / 2)
    k_value = 0.7071067811865476 * sqrt_n
    log_n_factorial = np.sum(np.log(np.arange(1, max(2, n + 1), 1, np.int64)))
    log_gamma = (
        0.5 * log_n_factorial
        - 0.5 * n * np.log(2.0)
        - n * np.log(k_value)
        - 0.25 * np.log(np.pi)
    )

    # now, all the values of the integrand are evaluated by making use of potentially
    # reduced computations that consider only some angles where the integrand exceeds
    # the machine precision
    log_eps = np.log(np.finfo(x.dtype).eps)
    one_over_sqrt_two_n = 1.0 / np.sqrt(2 * n)
    k_value_squared = k_value * k_value
    # for computing the interval bounds, the intersection of a quadratic cosine
    # polynomial with the logarithm of the machine precision are required
    # the constant part of the cosine value of this intersection value is pre-computed
    bound_value_part_two = (
        np.sqrt(n + 2.0 * (log_gamma - log_eps)) * one_over_sqrt_two_n
    )
    hermite_function = np.zeros_like(x)

    # each input point is evaluated separately
    # NOTE: this is probably only really fast for Numba-compiled code
    for iter_i, x_value in enumerate(x):
        # the x-value is made positive
        # NOTE: axis/point symmetry of the Hermite functions is exploited here, i.e.,
        #       for even orders, the function is symmetric around the y-axis and for
        #       odd orders, the function is anti-symmetric around the y-axis
        x_value_internal = abs(x_value)
        # the variable part of the bound cosine values is computed ...
        bound_value_part_one = x_value_internal * one_over_sqrt_two_n
        # ... followed by the values that the cosine has to yield for the intersections
        bound_value_left = bound_value_part_one - bound_value_part_two
        bound_value_right = bound_value_part_one + bound_value_part_two

        # if the left cosine value is within the bounds of -1 and 1, its arccos is
        # computed
        # otherwise, the bound angle is either pi (for < -1.0) or 0.0 (for > 1.0)
        if -1.0 <= bound_value_left <= 1.0:
            bound_value_left = np.arccos(bound_value_left)
        elif bound_value_left < -1.0:
            bound_value_left = np.pi
        else:
            bound_value_left = 0.0

        # the right bound is computed in almost the same way except for the fact that
        # the right cosine value cannot drop below 0.0 due to the absolute x-value
        # NOTE: the right bound value cannot drop below 0.0
        if bound_value_right <= 1.0:
            bound_value_right = np.arccos(bound_value_right)
        else:
            bound_value_right = 0.0

        # if the interval is degenerate, the integral is zero
        if bound_value_left == 0.0 and bound_value_right == 0.0:
            continue

        # now, the 2 sub-intervals for the integral are evaluated
        # the go from [bound_value_left, bound_value_right] for the negative angles and
        # [bound_value_right, bound_value_left] for the positive angles
        # NOTE: here, only the positive angles are considered because the target
        #       function's real part can be proven to be symmetric around the y-axis
        sub_integ_points = integ_points[
            np.where(
                np.logical_and(
                    integ_points >= bound_value_right,
                    integ_points <= bound_value_left,
                )
            )
        ]

        integrand_values_constant_log = log_gamma - 0.5 * np.square(x_value_internal)
        integrand_values_x_prefactor = 2 * x_value_internal * k_value
        # NOTE: the prefactor 2 exploits the symmetry of the target function, so only
        #       1 side has to be integrated
        hermite_function[iter_i] = 2.0 * trapezoid(
            y=np.exp(
                -1.0j * n * sub_integ_points
                + integrand_values_x_prefactor * np.exp(1.0j * sub_integ_points)
                - k_value_squared * np.exp(2.0j * sub_integ_points)
                + integrand_values_constant_log
            ).real,
            dx=delta_angle,
        )

        # finally, the symmetry of the Hermite functions is exploited
        # for even orders, the function is symmetric around the y-axis, so nothing has
        # to be done
        # for odd orders, the function is anti-symmetric around the y-axis, so the sign
        # has to be flipped
        if x_value < 0.0 and n % 2 == 1:
            hermite_function[iter_i] *= -1.0

    # finally, the hermite functions are properly scaled and returned
    return hermite_function / (2.0 * np.pi)
